// $Id: CharacterMap.java 10479 2009-07-10 09:51:07Z chris $

import java.util.*;
import java.io.*;
import java.awt.Color;
import org.faceless.pdf2.*;

/**
 * A more complex example which creates a Unicode Character Map, showing every
 * character in the font. This class will include characters in the U+10000
 * range if run under Java 1.5 or later, otherwise only U+0000 to U+FFFF are
 * supported (Search for NoSuchMethodError to see the differences).
 */
public class CharacterMap
{
    private static PDFStyle fontstyle, titlestyle, subtitlestyle, numberstyle, nonspacingstyle;
    private static PDFStyle boxglyph, boxmissing, boxundefined;

    public static void main(String[] args) throws IOException {
        PDF pdf = new PDF();
        PDFFont font = loadFont(args);
        initStyles(font);
        List bookmarks = pdf.getBookmarks();

        // Add some pages for each block to the document. There may
        // not be any characters defined for this block, in which case
        // addBlock() adds no pages and returns null.
        //
        for (int i=0;i<BLOCKS.length-1;i++) {
            Character.UnicodeBlock block;
            try {
                block = Character.UnicodeBlock.of(BLOCKS[i]);
            } catch (NoSuchMethodError e) {
                block = BLOCKS[i]>0xFFFF ? null : Character.UnicodeBlock.of((char)BLOCKS[i]);
            }
            if (block!=null) {
                String blockname = toTitleCase(block.toString());
                PDFPage page = createBlock(pdf, BLOCKS[i], BLOCKS[i+1], blockname, font);
                if (page!=null) {
                    bookmarks.add(new PDFBookmark(blockname, PDFAction.goTo(page)));
                }
            }
        }

        // Set some meta-information.
        pdf.setInfo("Title", "Character map for \""+font.getBaseName()+"\"");
        pdf.setInfo("Keywords", font.getBaseName()+", Unicode, character map, font");

        // Write the document to a file
        OutputStream out = new FileOutputStream("CharacterMap.pdf");
        pdf.render(out);
        out.close();
    }

    /**
     * Add the pages for the specified Unicode block to the document. This
     * may be zero or more pages (if there are no characters in this block).
     *
     * @param pdf the PDF
     * @param start the first Unicode codepoint in the block
     * @param end the last Unicode codepoint in the block
     * @param name the name of the block
     * @param font the Font to use
     * @return the first page in the block if one was created, or null otherwise.
     */
    private static PDFPage createBlock(PDF pdf, int start, int end, String name, PDFFont font) {
        boolean found = false;
        for (int i=start; !found && i<end; i++) {
            found = font.isDefined(i);
        }

        PDFPage firstpage = null;
        if (found) {
            System.out.print("* "+name);
            for (int pagestart=start; pagestart<end; pagestart+=128) {
                int pageend = end < pagestart+128 ? end : pagestart+128;
                found = false;
                for (int i=pagestart; !found && i<pageend; i++) {
                    found = font.isDefined(i);
                }
                if (found) {
                    PDFPage page = createPage(pdf, pagestart, pageend, name, font);
                    System.out.print(".");
                    if (firstpage==null) firstpage = page;
                }
            }
            System.out.println();
        }
        return firstpage;
    }

    /**
     * Create a page covering the specified range of characters.
     * @param pdf the PDF
     * @param start the first Unicode codepoint in the page
     * @param end the last Unicode codepoint in the page
     * @param name the name of the block
     * @param font the Font to use
     * @return the page
     */
    private static PDFPage createPage(PDF pdf, int start, int end, String name, PDFFont font) {
        PDFPage page = pdf.newPage("A4");
        page.setStyle(titlestyle);
        page.drawText(font.getBaseName(), page.getWidth()/2, page.getHeight()-50);
        page.setStyle(subtitlestyle);
        page.drawText(name+" ("+toHex(start)+" - "+toHex(end)+")", page.getWidth()/2, page.getHeight()-75);

        // Set the page to measure in percent, for simple layout
        page.setUnits(PDFPage.UNITS_PERCENT, PDFPage.ORIGIN_PAGETOP);
        int numrows = 16;
        float boxheight = 80 / numrows;         // 80%
        float boxwidth  = 80 / (128/numrows);

        for (int i=0; i<end-start; i++) {
            int row = i % numrows;
            int col = i / numrows;

            float left = 10 + (col*boxwidth);   // The 10 is 10% from the left/top
            float top =  10 + (row*boxheight);  

            int c = i+start;
            String text = c<0x10000?""+(char)c:""+(char)((c>>10)+0xD7C0)+(char)((c&0x3FF)+0xDC00);
            boolean infont = font.isDefined(c);
            boolean valid;
            try {
                valid = Character.getType(c)!=Character.UNASSIGNED;
            } catch (NoSuchMethodError e) {
                valid = Character.getType((char)c)!=Character.UNASSIGNED;
            }

            boolean nonspacingmark = false;
            if (infont && valid && fontstyle.getTextLength(text)==0) {
                text = "\u00A0"+text+"\u00A0";
                nonspacingmark = true;
            }

            // If the character is defined, draw the glyph and draw a glyph box
            // otherwise set to gray box and draw the rectangle.
            if (infont) {
                page.setStyle(fontstyle);
                page.drawText(text, left+(boxwidth*0.5f), top+(boxheight*0.6f));
                page.setStyle(boxglyph);
            } else if (valid) {
                page.setStyle(boxmissing);
            } else {
                page.setStyle(boxundefined);
            }
            page.drawRectangle(left, top, left+boxwidth, top+boxheight);
            if (valid) {
                page.setStyle(numberstyle);         // Add the Unicode point to the box
                page.drawText(toHex(c), left+(boxwidth*0.9f), top+(boxheight*0.9f));
            }

            // Draw the dotted-circle to indicate where the character would be in a
            // diacritic or combining mark. Not 100% accurate, especially for
            // right-to-left combining marks (eg. arabic vowel sounds), but good enough.
            // 
            if (nonspacingmark && nonspacingstyle!=null) {
                page.setStyle(nonspacingstyle);
                page.drawCircle(left+(boxwidth*0.45f), top+(boxheight*0.52f), boxwidth/20);
            }
        }
        return page;
    }

    /**
     * Initialise the styles
     */
    private static void initStyles(PDFFont font) {
        // "fontstyle" is used to actually render the glyphs.
        fontstyle = new PDFStyle();
        fontstyle.setFillColor(Color.black);
        fontstyle.setFont(font, 12);
        fontstyle.setTextAlign(PDFStyle.TEXTALIGN_CENTER);

        // "titlestyle"
        titlestyle = new PDFStyle();
        titlestyle.setFont(new StandardFont(StandardFont.TIMES), 20);
        titlestyle.setFillColor(Color.black);
        titlestyle.setTextAlign(PDFStyle.TEXTALIGN_CENTER);
 
        // "subtitlestyle"
        subtitlestyle = new PDFStyle(titlestyle);
        subtitlestyle.setFont(new StandardFont(StandardFont.HELVETICA), 15);

        // "numberstyle" is for the "U+xxxx" notes in each character box
        numberstyle = new PDFStyle();
        numberstyle.setFont(new StandardFont(StandardFont.HELVETICAOBLIQUE), 6);
        numberstyle.setFillColor(Color.black);
        numberstyle.setTextAlign(PDFStyle.TEXTALIGN_RIGHT);

        // "boxglyph" is to surround glyphs that actually exist
        boxglyph = new PDFStyle();
        boxglyph.setLineColor(Color.black);

        // "boxmissing" is to surround glyphs that are valid but not in the font
        boxmissing = (PDFStyle)boxglyph.clone();
        boxmissing.setFillColor(new Color(220, 220, 220));

        // "boxinvalid" is to surround glyphs that aren't defined in Unicode.
        boxundefined = (PDFStyle)boxglyph.clone();
        boxundefined.setFillColor(Color.black);

        // "nonspacingstyle" is to draw the circle that indicates a non-spacing mark
        // or diacritic.
        //
        // Set this to null to not draw a circle - down to your preference really.
        //
        nonspacingstyle = new PDFStyle();
        nonspacingstyle.setLineColor(new Color(200, 200, 200));
        nonspacingstyle.setLineWeighting(0.5f);
        nonspacingstyle.setLineDash(1, 1, 0);
    }

    /**
     * Load the font.
     */
    private static PDFFont loadFont(String[] args) throws IOException {
        PDFFont font = null;
        String ttfname = null, afmname = null, pfbname = null, stdname = null;

        if (args.length>0) {
            String fontname = args[0];
            if (new File(fontname).canRead()) {
                if (fontname.endsWith(".afm")) {
                    InputStream afm = new FileInputStream(fontname);
                    InputStream pfb = args.length==1 ? null : new FileInputStream(args[1]);
                    font = new Type1Font(afm, pfb);
                } else {
                    font = new OpenTypeFont(new FileInputStream(fontname), 2);
                }
            } else {
                for (int i=0;font==null && i<STDFONT.length;i++) {
                    if (STDFONT[i].equalsIgnoreCase(fontname)) {
                        font = new StandardFont(i);
                    }
                }
                for (int i=0;font==null && i<STDCJKFONT.length;i++) {
                    if (STDCJKFONT[i].equalsIgnoreCase(fontname)) {
                        font = new StandardCJKFont(i, 0);
                    }
                }
            }
        }
        if (font==null) {
            System.out.println("Usage: CharacterMap [Times-Roman | Times-Bold | Times-Italic | Times-BoldItalic");
            System.out.println("                  Helvetica | Helvetica-Bold | Helvetica-Oblique | Helvetica-BoldOblique |");
            System.out.println("                  Courier | Courier-Bold | Courier-Oblique | Courier-BoldOblique |");
            System.out.println("                  Symbol | ZapfDingBats | STSong | MSung | MHei | HeiSeiMin |");
            System.out.println("                  HeiSeiKakuGo | HYGothic | HYSMyeongJo | <fontfile> ] [<fontfile.pfb>]");
            System.exit(0);
        }

        return font;
    }

    //-----------------------------------------------------------------------
    // The start of the Unicode code blocks. Lifted gratefully from
    // the source to java.lang.Character.UnicodeBlock.
    //
    private static final int[] BLOCKS = { 0x0000, 0x0080, 0x0100, 0x0180, 0x0250, 0x02B0, 0x0300, 0x0370, 0x0400, 0x0500, 0x0530, 0x0590, 0x0600, 0x0700, 0x0750, 0x0780, 0x07C0, 0x0900, 0x0980, 0x0A00, 0x0A80, 0x0B00, 0x0B80, 0x0C00, 0x0C80, 0x0D00, 0x0D80, 0x0E00, 0x0E80, 0x0F00, 0x1000, 0x10A0, 0x1100, 0x1200, 0x1380, 0x13A0, 0x1400, 0x1680, 0x16A0, 0x1700, 0x1720, 0x1740, 0x1760, 0x1780, 0x1800, 0x18B0, 0x1900, 0x1950, 0x1980, 0x19E0, 0x1A00, 0x1D00, 0x1D80, 0x1E00, 0x1F00, 0x2000, 0x2070, 0x20A0, 0x20D0, 0x2100, 0x2150, 0x2190, 0x2200, 0x2300, 0x2400, 0x2440, 0x2460, 0x2500, 0x2580, 0x25A0, 0x2600, 0x2700, 0x27C0, 0x27F0, 0x2800, 0x2900, 0x2980, 0x2A00, 0x2B00, 0x2C00, 0x2C60, 0x2C80, 0x2D00, 0x2D30, 0x2D80, 0x2E00, 0x2E80, 0x2F00, 0x2FE0, 0x2FF0, 0x3000, 0x3040, 0x30A0, 0x3100, 0x3130, 0x3190, 0x31A0, 0x31C0, 0x31F0, 0x3200, 0x3300, 0x3400, 0x4DC0, 0x4E00, 0xA000, 0xA490, 0xA4D0, 0xAC00, 0xD7B0, 0xD800, 0xDB80, 0xDC00, 0xE000, 0xF900, 0xFB00, 0xFB50, 0xFE00, 0xFE10, 0xFE20, 0xFE30, 0xFE50, 0xFE70, 0xFF00, 0xFFF0, 0x10000, 0x10080, 0x10100, 0x10140, 0x10300, 0x10330, 0x10350, 0x10380, 0x103A0, 0x10400, 0x10450, 0x10480, 0x104B0, 0x10800, 0x10840, 0x1D000, 0x1D100, 0x1D200, 0x1D300, 0x1D360, 0x1D400, 0x1D800, 0x20000, 0x2A6E0, 0x2F800, 0x2FA20, 0xE0000, 0xE0080, 0xE0100, 0xE01F0, 0xF0000, 0x100000 };

    static final String[] STDFONT = { "Times-Roman", "Times-Bold", "Times-Italic", "Times-BoldItalic", "Helvetica", "Helvetica-Bold", "Helvetica-Oblique", "Helvetica-BoldOblique", "Courier", "Courier-Bold", "Courier-Oblique", "Courier-BoldOblique", "Symbol", "ZapfDingbats" };
    static final String[] STDCJKFONT = { "STSong", "MSung", "MHei", "HeiSeiMin", "HeiseiKakuGo", "HYGothic", "HYSMyeongJo" };

    /**
     * Given a character, eg 'A', format it as "U+0041"
     */
    private static String toHex(int c) {
        String s = Integer.toHexString(c).toUpperCase();
        while (s.length()<4) s="0"+s;
        return "U+"+s;
    }

    /**
     * Take the block name and make it nicer
     */
    private static String toTitleCase(String s) {
        StringBuffer sb = new StringBuffer(s.length());
        boolean toupper = true;
        int i=0;
        if (s.startsWith("CJK_") || s.startsWith("IPA_")) {
            sb.append(s.substring(0, 3));
            i=sb.length();
        }
        while (i<s.length()) {
            char c = s.charAt(i);
            if (c=='_') c=' ';
            sb.append(toupper ? Character.toUpperCase(c) : Character.toLowerCase(c));
            toupper = !Character.isUpperCase(c);
            i++;
        }
        return sb.toString();
    }
}
