// $Id: DiskCache.java 14646 2011-12-14 09:28:46Z chris $

package org.faceless.pdf2;

import java.io.*;
import org.faceless.util.BackedInputStream;

/**
 * A general purpose {@link Cache} which optionally writes closed streams
 * to files on disk. This class has been tested on UNIX and Windows under
 * multi-threaded environments, and we see no reason why it shouldn't work
 * under other platforms as well. Here's a simple example:
 * <pre class="example">
 *   Cache cache = new DiskCache(System.getProperty("java.io.tmpdir")+"/pdftemp.", 8192);
 *   PDF.setCache(cache);
 * </pre>
 * @see PDF#setCache
 * @since 2.2.2
 */
public class DiskCache implements Cache
{
    private volatile static int next;   // The next number. Static just in case two DiskCaches
                                        // are created on the same directory.

    private final String prefix;              // The directory to store the cached files in
    private final int threshold;              // The smallest stream to write to disk
    boolean debug;

    /**
     * Create a new DiskCache
     *
     * @param prefix The prefix to begin the cached filenames with. For instance, if
     * prefix was "/tmp/cache/mycache.", the cache files would be "/tmp/cache/mycache.0",
     * "/tmp/cache/mycache.1" etc.
     *
     * @param threshold the minimum number of bytes that are considered worth caching to
     * disk, or 0 to prevent caching altogether
     */
    public DiskCache(String prefix, int threshold) {
        this.prefix = prefix;
        this.threshold = threshold==0 || prefix==null ? Integer.MAX_VALUE : threshold;
    }


    public Cache.Entry newEntry(int size) {
        return new DiskCacheEntry(size);
    }

    /**
     * Return the next filename to be used in the cache (of the form "prefixnnn")
     */
    private synchronized File nextFileName() {
        return prefix==null ? null : new File(prefix+(next++)) {
            protected void finalize() {
                if (!debug) delete();
            }
        };
    }


    private class DiskCacheEntry extends OutputStream implements Cache.Entry {
        private OutputStream out;
        private File file;                      // null if data held in memory
        private int count;
        private boolean closed;

        public DiskCacheEntry(int size) {
            if (prefix!=null && size>=threshold) {
                try {
                    toDisk();
                } catch (IOException e) {
                    throw toRuntimeException(e);
                }
            } else {
                out = new MyByteArrayOutputStream(size);
            }
        }

        public OutputStream getOutputStream() {
            if (closed) {
                throw new IllegalStateException("Can't write to closed stream \""+file+"\"");
            }
            return this;
        }

        public BackedInputStream getInputStream() {
            if (file!=null) {
                try {
                    if (!closed) out.flush();
                    return BackedInputStream.getInstance(file);
                } catch (IOException e) {
                    throw toRuntimeException(e);
                }
            } else {
                return BackedInputStream.getInstance(((MyByteArrayOutputStream)out).buf, 0, count);
            }
        }

        public int size() {
            return count;
        }

        public void writeTo(OutputStream stream) throws IOException {
            if (out instanceof MyByteArrayOutputStream) {
                stream.write(((MyByteArrayOutputStream)out).buf, 0, count);
            } else {
                BackedInputStream in = getInputStream();
                in.writeTo(stream);
                in.close();
            }
        }

        public Cache.Entry cloneEntry() {
            if (closed) {
                return this;
            } else {
                DiskCacheEntry copy = new DiskCacheEntry(count);
                try {
                    OutputStream out = null;
                    try {
                        out = copy.getOutputStream();
                        writeTo(out);
                    } finally {
                        if (out != null) {
                            out.close();
                        }
                    }
                } catch (IOException e) {
                    throw toRuntimeException(e);
                }
                return copy;
            }
        }

        public void close() {
            if (!closed) {
                try {
                    out.close();
                    if (file!=null) {
                        out = null;
                    }
                } catch (IOException e) {
                    throw toRuntimeException(e);
                }
                closed = true;
            }
        }

        public String toString() {
            return "<filecache size="+size()+" file="+file+">";
        }


        //-----  OutputStream methods from here on ----

        public void write(int b) throws IOException {
            out.write(b);
            count++;
            if (prefix!=null && file==null && count>threshold) {
                toDisk();
            }
        }

        public void write(byte[] b, int off, int len) throws IOException {
            if (prefix!=null && file==null && count+len>threshold) {
                toDisk();
            }
            out.write(b, off,len);
            count += len;
        }

        public void write(byte[] b) throws IOException {
            write(b, 0, b.length);
        }

        private void toDisk() throws IOException {
            file = nextFileName();
            if (!debug) {
                file.deleteOnExit();
            }
            byte[] buf = out==null ? null : ((MyByteArrayOutputStream)out).buf;
            out = new BufferedOutputStream(new FileOutputStream(file), 128);
            if (buf!=null) {
                out.write(buf, 0, count);
            }
        }
    }

    private static final class MyByteArrayOutputStream extends OutputStream {
        private int count;
        byte[] buf;

        MyByteArrayOutputStream(int size) {
            buf = new byte[size==0 ? 16 : size];
        }

        public void write(int b) {
            if (count==buf.length) {
                extend(count+8);
            }
            buf[count++] = (byte)b;
        }

        public void write(byte[] b, int off, int len) {
            if (len==0) {
                return;
            } else if (off<0 || len<0 || off+len > b.length) {
                throw new IndexOutOfBoundsException(off+"+"+len+" outside "+b.length);
            }
            if (count+len > buf.length) {
                extend(count+len);
            }
            System.arraycopy(b, off, buf, count, len);
            count += len;
        }

        public void write(byte[] b) {
            write(b, 0, b.length);
        }

        /**
         * Increase the capacity of the internal buffer
         */
        private void extend(int newlen) {
            if (newlen > buf.length) {
                if (newlen < buf.length+(buf.length>>1)) {
                    newlen = buf.length+(buf.length>>1);
                }
                byte[] newbuf = new byte[newlen];
                System.arraycopy(buf, 0, newbuf, 0, count);
                buf = newbuf;
            }
        }
    }

    private static final RuntimeException toRuntimeException(Exception e) {
        if (e instanceof RuntimeException) {
            return (RuntimeException)e;
        } else {
            try {
                Class.forName("java.lang.StackTraceElement");   // Test for Java 1.4
                return new RuntimeException(e);
            } catch (Throwable e2) {
                return new RuntimeException(e.toString());
            }
        }
    }
}
