// $Id: AppleSupport.java 14034 2011-10-13 16:44:21Z mike $

package org.faceless.pdf2.viewer2.feature;

import org.faceless.pdf2.viewer2.*;
import org.faceless.pdf2.PDFPage;
import javax.swing.*;

/**
 * <p>
 * This feature can be added to the viewer on Mac OS X for a more "Mac-like"
 * interface. Some of the menu items will be moved around to match OS X
 * interface guidelines, and rotate and swipe gestures can be used to 
 * rotate and switch between pages. Adding this class on a non-OS X system
 * has no effect.
 * </p>
 *
 * <div class="initparams">
 * The following <a href="../doc-files/initparams.html">initialization parameters</a> can be specified to configure this feature.
 * <table>
 * <tr><th>dockIcon</th><td><code>true</code> or <code>false</code>, for {@link #setDockIcon}</td></tr>
 * <tr><th>moveMenus</th><td><code>true</code> or <code>false</code>, for {@link #setMoveMenus}</td></tr>
 * </table>
 * </div>
 *
 * <span class="featurename">The name of this feature is <span class="featureactualname">AppleSupport</span></span>
 * <p><i>This code is copyright the Big Faceless Organization. You're welcome to use, modify and distribute it in any form in your own projects, provided those projects continue to make use of the Big Faceless PDF library.</i></p>
 * @since 2.11.1
 */
public class AppleSupport extends ViewerFeature {

    private boolean menus, dockicon;

    public AppleSupport() {
        super("AppleSupport");
        dockicon = true;
    }

    /**
     * Set whether to move the menus from the application window to the System
     * menu bar.
     * @param menus whether to move the menus or not
     * @since 2.11.6
     */
    public void setMoveMenus(boolean menus) {
        this.menus = menus;
    }

    /**
     * Set whether to set the dock icon. The default is true.
     * @param dockicon whether to customize the dock icon
     * @since 2.11.6
     */
    public void setDockIcon(boolean dockicon) {
        this.dockicon = dockicon;
    }

    public void initialize(final PDFViewer viewer) {
        super.initialize(viewer);
        if (Util.isLAFAqua()) {
            String val = getFeatureProperty(viewer, "moveMenus");
            if (val!=null) {
                setMoveMenus("true".equals(val));
            }
            val = getFeatureProperty(viewer, "dockIcon");
            if (val!=null) {
                setDockIcon("true".equals(val));
            }

            try {
                AppleSupportImpl.initialize(viewer, menus, dockicon);
            } catch (Throwable e) {
            }
        }
    }
}

class AppleSupportImpl {
    static void initialize(final PDFViewer viewer, boolean menus, boolean dockicon) {
        com.apple.eawt.Application app = com.apple.eawt.Application.getApplication();
        if (menus && viewer.hasFeature(Menus.getInstance())) {
            System.setProperty("apple.laf.useScreenMenuBar", "true");
            JMenu menu = viewer.getMenu("Window");
            if (menu!=null) {
                for (int i=0;i<menu.getMenuComponentCount();i++) {
                    final JMenuItem item = menu.getItem(i);
                    if (item.getText().equals(UIManager.get("PDFViewer.About"))) {
                        menu.remove(item);
                        app.setAboutHandler(new com.apple.eawt.AboutHandler() {
                            public void handleAbout(com.apple.eawt.AppEvent.AboutEvent event) {
                                item.doClick();
                            }
                        });
                    }
                }
            }
            menu = viewer.getMenu("File");
            for (int i=0;i<menu.getMenuComponentCount();i++) {
                final JMenuItem item = menu.getItem(i);
                if (item.getText().equals(UIManager.get("PDFViewer.Quit"))) {
                    menu.remove(item);
                    app.setQuitHandler(new com.apple.eawt.QuitHandler() {
                        public void handleQuitRequestWith(com.apple.eawt.AppEvent.QuitEvent event, com.apple.eawt.QuitResponse response) {
                            item.doClick();
                        }
                    });
                }
            }
        }

        // GestureHandler for magnify, rotate and swipe.
        // Magnification turned off - we can't do it realtime, it seems to jump around a bit
        // Rotate works
        // Swipe - left and right seem to be reversed (up and down aren't used)
        //
        final com.apple.eawt.event.GestureListener gesturelistener = new com.apple.eawt.event.GestureAdapter() {
            private double rotate, magnify;
            public void gestureBegan(com.apple.eawt.event.GesturePhaseEvent e) {
                rotate = magnify = 0;
            }
            public void gestureEnded(com.apple.eawt.event.GesturePhaseEvent e) {
                while (rotate < 0) rotate += 360;
                rotate = rotate % 360;
                DocumentPanel docpanel = viewer.getActiveDocumentPanel();
                PDFPage page = docpanel.getPage();
//                System.out.println("END: rot="+rotate+" mag="+magnify+" pagerot="+page.getPageOrientation());
                if (rotate > 315 || rotate < 45) {
                    if (false && magnify != 0) {
                        double zoom = docpanel.getZoom();
                        zoom *= magnify;
                        zoom = Math.min(Math.max(zoom, 0.125), 64);
                        docpanel.setZoom((float)zoom);
                    }
                } else if (rotate >= 225) {             // Rotate 90 degrees clockwise
                    page.setPageOrientation(page.getPageOrientation() + 90);
                } else if (rotate >= 135) {             // Rorate 180 degrees
                    page.setPageOrientation(page.getPageOrientation() + 180);
                } else if (rotate >= 45) {              // Rotate 90 counter-clockwise
                    page.setPageOrientation(page.getPageOrientation() - 90);
                }
                rotate = magnify = 0;
            }
            public void magnify(com.apple.eawt.event.MagnificationEvent e) {
                magnify += e.getMagnification();
            }
            public void rotate(com.apple.eawt.event.RotationEvent e) {
                rotate += e.getRotation();
            }
            public void swipedRight(com.apple.eawt.event.SwipeEvent e) {
                DocumentPanel docpanel = viewer.getActiveDocumentPanel();
                int page = docpanel.getViewport().getRenderingPage().getPageNumber();
                if (page > 1) {
                    page--;
                    docpanel.setPageNumber(page - 1);
                }
            }
            public void swipedLeft(com.apple.eawt.event.SwipeEvent e) {
                DocumentPanel docpanel = viewer.getActiveDocumentPanel();
                int page = docpanel.getViewport().getRenderingPage().getPageNumber();
                if (page < docpanel.getPDF().getNumberOfPages()) {
                    page++;
                    docpanel.setPageNumber(page - 1);
                }
            }
        };

        viewer.addDocumentPanelListener(new DocumentPanelListener() {
            public void documentUpdated(DocumentPanelEvent event) {
                if (event.getType() == "activated") {
                    com.apple.eawt.event.GestureUtilities.addGestureListenerTo(event.getDocumentPanel(), gesturelistener);
                } else if (event.getType() == "deactivated") {
                    com.apple.eawt.event.GestureUtilities.removeGestureListenerFrom(event.getDocumentPanel(), gesturelistener);
                }
            }
        });

        if (dockicon) {
            try {
                app.setDockIconImage(new ImageIcon(PDFViewer.class.getResource("resources/bfodockicon.png")).getImage());
            } catch (Throwable e) {}
        }
    }

}
